#include "Controls.h"

// Drive Control Bindings
const int kDriveToggleModes = 1; // a button
const int kDriveSwapCamera = 2; // B button

const int kDriveTankLeftAxis = 1; // left stick, y axis
const int kDriveTankRightAxis = 5; // right stick, y axis

const int kDriveArcadeForwardAxis = 1; // left stick, y axis
const int kDriveArcadeTurnAxis = 4; // right stick, x axis

const int kDriveLeftTurboAxis = 2; // left trigger
const int kDriveRightTurboAxis = 3; // right trigger

const int kDriveLeftSlow = 5; // left bumper
const int kDriveRightSlow = 6; //right bumper

const int kCalibrateGyro = 4; // Y button

// Aux Control Bindings
const int kAuxControlPanelSpin = 2;

// Tunable constants
const double kSlowMultiplier = .3; // .3
const double kNormalMultiplier = .6; // .5
const double kTurboMultiplier = .7; // .7
const double kMaxDriveSpeed = .8;
const double kTurboPreRampDuration = 1; // wait 1 second after starting turbo to speed it up more

const double kTurnMultiplier = .65;

const double kArcadeTurnDeadzone = .1;

// Xbox button maps
const int kAButton = 1; // GetRawButton() give bool
const int kBButton = 2; // GetRawButton() give bool
const int kXButton = 3; // GetRawButton() give bool
const int kYButton = 4; // GetRawButton() give bool
const int kLeftBumper = 5; // GetRawButton() give bool
const int kRightBumper = 6; // GetRawButton() give bool
const int kBackButton = 7; // GetRawButton() give bool
const int kStartButton = 8; // GetRawButton() give bool
const int kLeftTrigger = 2; // GetRawAxis() give float
const int kRightTrigger = 3; // GetRawAxis() give float
const int kLeftDial = 1; // GetRawAxis() give float
const int kRightDial = 5; // GetRawAxis() give float
const int kDPad = 0; // GetPOV() give float
// Switches
const int kSwitchOne = 1; // GetRawButton() give bool
const int kSwitchTwo = 2; // GetRawButton() give bool
const int kSwitchThree = 3; // GetRawButton() give bool
const int kSwitchFour = 4; // GetRawButton() give bool
const int kSwitchFive = 5; // GetRawButton() give bool
const int kSwitchSix = 6; // GetRawButton() give bool
const int kSwitchSeven = 7; // GetRawButton() give bool
const int kSwitchEight = 8; // GetRawButton() give bool
const int kSwitchNine = 9; // GetRawButton() give bool

Controls::Controls(Drive* drive, GamePiece* gamePiece, Hang* hang, ControlPanel* controlPanel) {
    this->drive = drive;
    this->gamePiece = gamePiece;
    this->hang = hang;
    this->controlPanel = controlPanel;
}

void Controls::process() {
    processDriver();
    processAux();
    brokenSensor();
}

void Controls::processDriver() {
    bool toggleDriveMode = controllerDriver.GetRawButton(kDriveToggleModes);

    double tankLeft = -controllerDriver.GetRawAxis(kDriveTankLeftAxis);
    double tankRight = -controllerDriver.GetRawAxis(kDriveTankRightAxis);

    double arcadeForward = -controllerDriver.GetRawAxis(kDriveArcadeForwardAxis);
    double arcadeTurn = controllerDriver.GetRawAxis(kDriveArcadeTurnAxis);

    bool turboLeft = controllerDriver.GetRawAxis(kDriveLeftTurboAxis) > .75;
    bool turboRight = controllerDriver.GetRawAxis(kDriveRightTurboAxis) > .75;

    bool slowLeft = controllerDriver.GetRawButton(kDriveLeftSlow);
    bool slowRight = controllerDriver.GetRawButton(kDriveRightSlow);

    bool calibrateGyro = controllerDriver.GetRawButton(kCalibrateGyro);

    double hangSlowForward = controllerDriver.GetPOV();

    bool swapCamera = controllerDriver.GetRawButtonPressed(kDriveSwapCamera);

    if(calibrateGyro) {
        drive->calibrateGyro();
    }

    if(swapCamera) {
        viewingLimelight = !viewingLimelight;
    }

    if(toggleDriveMode == true) {
        if(wasDriveModeToggled == false) {
            isTankMode = !isTankMode;
        }
    }

    wasDriveModeToggled = toggleDriveMode;

    double finalLeft = 0;
    double finalRight = 0;
    
    if(isTankMode) { // tank drive
        double tankOutputLeft = getPower(tankLeft, turboLeft, slowLeft);
        double tankOutputRight = getPower(tankRight, turboRight, slowRight);
       
        finalLeft = tankOutputLeft;
        finalRight = tankOutputRight;
    }
    else { // arcade drive
        double arcadeOutputLeft = arcadeForward;
        double arcadeOutputRight = arcadeForward;

        if(arcadeForward < -.05) {
            arcadeTurn *= -1;
        }
        
        if(arcadeTurn < -kArcadeTurnDeadzone) { /// turn left
            if(arcadeOutputLeft > 0) {// if driving forwards
                arcadeOutputLeft -= fabs(arcadeTurn) * kTurnMultiplier;
            }
            else if(arcadeOutputLeft < 0) { // if driving backwards
                arcadeOutputLeft += fabs(arcadeTurn) * kTurnMultiplier;
            }
        }
        else if(arcadeTurn > kArcadeTurnDeadzone) { // turn right
            if(arcadeOutputRight > 0) {// if driving forwards
                arcadeOutputRight -= arcadeTurn * kTurnMultiplier;
            }
            else if(arcadeOutputRight < 0) { // if driving backwards
                arcadeOutputRight += arcadeTurn * kTurnMultiplier;
            }
        }
        
        if(abs(arcadeTurn) > kArcadeTurnDeadzone) {
            if(abs(arcadeForward) < .1) {
                arcadeOutputLeft = arcadeTurn;
                arcadeOutputRight = -arcadeTurn;
            }
        }

        if(turboLeft && turboRight) {
            turboHeld.Start();
        }
        else {
            turboHeld.Reset();
            turboHeld.Stop();
        }

        arcadeOutputLeft = getPower(arcadeOutputLeft, turboLeft, slowLeft);
        arcadeOutputRight = getPower(arcadeOutputRight, turboRight, slowRight);

        finalLeft = arcadeOutputLeft;
        finalRight = arcadeOutputRight;
    }

    if(fabs(finalLeft) < .1 && fabs(finalRight) < .1) {
        if(hangSlowForward == 0) {
            finalLeft = .2;
            finalRight = .2;
        }
    } 

    drive->setSpeed(finalLeft, finalRight);
}


void Controls::processAux() {
    // These are my Aux variables
     bool intakeLT;
     bool shooterHelixRB;
     float rotateTurretLD;
     float shooterSpeedRD;
     bool retractHangA;
     bool hoodWallB;
     bool hoodInitLineX;
     bool extendHangY;
     float controlPanelDPad;
     bool outtakeLB;
     float shootBallRT;
     bool hoodPanelSB;
     bool positionRotation;
     bool conrtolPanelManual;
     double shooterManualSpeed;

     // Sets true if the button is being pressed down
     intakeLT = controllerAux.GetRawAxis(kLeftTrigger) > .4;
     shooterHelixRB = controllerAux.GetRawButton(kRightBumper);
     rotateTurretLD = controllerAux.GetRawAxis(kLeftDial);
     shooterSpeedRD = controllerAux.GetRawAxis(kRightDial);
     retractHangA = controllerAux.GetRawButton(kAButton);
     hoodWallB = controllerAux.GetRawButton(kBButton);
     hoodInitLineX = controllerAux.GetRawButton(kXButton);
     extendHangY = controllerAux.GetRawButton(kYButton);
     controlPanelDPad = controllerAux.GetPOV(kDPad);
     outtakeLB = controllerAux.GetRawButton(kLeftBumper);
     shootBallRT = controllerAux.GetRawAxis(kRightTrigger);
     hoodPanelSB = controllerAux.GetRawButton(kStartButton);
     shooterManualSpeed = -controllerAux.GetRawAxis(1);
     positionRotation = controllerAux2.GetRawButton(kSwitchSix);
     conrtolPanelManual = controllerAux2.GetRawButton(kSwitchOne);
     

     
     // Turns on Intake if Outake is off
     if(intakeLT == true && outtakeLB == false){ // Only runs intake if isnt outtaking
         gamePiece->setIntakeState(GamePiece::INTAKE);
     }
     // Turns on Outtake, turns it off if neither outake or intake are on
     else if(outtakeLB == true && intakeLT == false){ // Only runs outtake if isnt intakking
         gamePiece->setIntakeState(GamePiece::OUTTAKE);
     }
     else{ // Runs neither if both are pressed or if neither are pressed
         gamePiece->setIntakeState(GamePiece::NOTTAKE);
     }
     // Sets the shooter speed to the speed on the dial
     gamePiece->setShooterSpeed(shooterSpeedRD);
     // Turns on Shooter and Helix if not intaking
     if(shooterHelixRB == true){
         gamePiece->setShooterPreheat(true);
     }
     else{
         gamePiece->setShooterPreheat(false);
     }
     
     // Actually shoots the ball if warmed up
     if(shootBallRT >= .6) {
         gamePiece->activateShooter(true);
     }
     else{
         gamePiece->activateShooter(false);
     }
     /* Testish for rotate turret
     // Spins the turret to the dial position
     if turretPosition != rotateTurretLD{
         shooter->(rotateTurretLD) // Need function to spin turret
     } */
     // Starts control panel if button pressed, if switch is true does rotation, if it isnt does position
     if (conrtolPanelManual == true){ // Checks if the manual switch is activated, turns into a press and hold
         if(controlPanelDPad == 0){
             if(positionRotation == true){
                 controlPanel->manualDriveFast();
             }
             else{
                 controlPanel->manualDriveSlow();
             }
         }
         else{
             controlPanel->stop();
         }
     }
     else{ // If no manual switch run control panel like normal
         if(controlPanelDPad == 0){ // If the control panel button was pressed
            if(positionRotation == true){ // If the control switch is in rotation mode do rotation control
             controlPanel->rotationControl();
            }
            else{ // If the control switch isnt in rotaition mode do position control
             controlPanel->positionControl();
            }
        }
        else {
            controlPanel->stop();
        }
     }
     if(extendHangY == true && retractHangA == false){ // Extends hang if isnt retracting
         hang->move(Hang::UP);
     }
     // Retracts hang if button pressed
     else if(retractHangA == true && extendHangY == false){ // Retracts hang if isnt extending
         hang->move(Hang::DOWN);
     }
     else{ // If both pressed or neither pressed it does nothing
         hang->move(Hang::STOP);
     }
     // Turns shooter turret
     if(rotateTurretLD == true){ 
     }
     // Sets shooter to initiation line
      if(hoodInitLineX == true){
        gamePiece->hood(Shooter::LINE);
     }
     // Sets shooter to Control Panel
     if(hoodPanelSB == true){
         gamePiece->hood(Shooter::PANEL);
     }
     // Sets shooter to wall
     if(hoodWallB == true){
         gamePiece->hood(Shooter::PORT);
     }
     

     if(shooterManualSpeed > .1) { // up
        if(desiredShooterSpeed < 1) {
            desiredShooterSpeed += .01;
        }
     }
     else if(shooterManualSpeed < -.1) { // down
        if(desiredShooterSpeed > 0) {
            desiredShooterSpeed -= .01;
        }
     }

     gamePiece->shooter->setShooterSpeed(desiredShooterSpeed);
}

double Controls::getPower(double speed, bool turbo, bool slow) {
    if(slow) {
        speed *= kSlowMultiplier;
    }
    else if(!turbo) {
        speed *= kNormalMultiplier;
    }
    else {
        speed *= kTurboMultiplier;
        if(speed < kMaxDriveSpeed && turboHeld.Get() > kTurboPreRampDuration) {
            speed += (turboHeld.Get() - kTurboPreRampDuration) * .03;
        }
    }

    return speed;
}
void Controls::brokenSensor() {
    bool intakePivotSensor;
    bool intakeBallCounterSensor;
    bool storageTransitionSensor;
    bool shooterBallCounterSensor;
    // bool hangZeroSensor;
    bool hangRetractSensor;
    bool shooterManualSpeed;
    bool driveEncBroke;

    intakePivotSensor = controllerAux2.GetRawButton(kSwitchTwo);
    intakeBallCounterSensor = controllerAux2.GetRawButton(kSwitchThree);
    storageTransitionSensor = controllerAux2.GetRawButton(kSwitchFour);
    shooterBallCounterSensor = controllerAux2.GetRawButton(kSwitchFive);
    // hangZeroSensor = controllerAux2.GetRawButton(kSwitchSeven);
    hangRetractSensor = controllerAux2.GetRawButton(kSwitchNine);
    shooterManualSpeed = controllerAux2.GetRawButton(kSwitchSeven);
    driveEncBroke = controllerAux2.GetRawButton(8);

    drive->setBackdriveBroken(driveEncBroke);
    

    // If each broken switch is flipped it will run a function with true, otherwise it runs said function with false
    if (intakePivotSensor == true){
        gamePiece->pivotSensorBroken(true);
    }
    else{
        gamePiece->pivotSensorBroken(false);
    }
    if (intakeBallCounterSensor == true){
        gamePiece->setIntakeCounterBroken(true);
    }
    else{
        gamePiece->setIntakeCounterBroken(false);
    }
    if (storageTransitionSensor == true){
        gamePiece->setTransitionCounterBroken(true);
    }
    else{
        gamePiece->setTransitionCounterBroken(false);
    }
    if (shooterBallCounterSensor == true){
        gamePiece->setShooterBallCounterBroken(true);
    }
    else{
        gamePiece->setShooterBallCounterBroken(false);
    }
    // if (hangZeroSensor == true){
    //     hang->zeroSensorBroken(true);
    // }
    // else{
    //     hang->zeroSensorBroken(false);
    // }

    if(hangRetractSensor) {
        hang->enableSlowRetract(false);
    }
    else {
        hang->enableSlowRetract(true);
    }

    gamePiece->shooter->enableManualControl(shooterManualSpeed);
}

void Controls::debug(Feedback* feedback) {
    feedback->sendString("controls", "drive style", isTankMode ? "tank" : "arcade");
    feedback->sendDouble("thunderdashboard", "drivearcade", isTankMode ? 0 : 1);
    feedback->sendDouble("thunderdashboard", "frontcamera", viewingLimelight ? 0 : 1);
}
