#include "Drive.h"
#include "Utils.h"
#include <frc/smartdashboard/SmartDashboard.h>

const int kTicksPerInch = 109;

const double kAutoStraightAcceptableRange = 2.0; // inches
const double kAutoPivotAcceptableRange = 10; // degrees

const double kAutoPivotSlowdownRange = 40; // degrees
const double kAutoPivotSlowdownSpeed = .1;

const double kAutoStraightSlowdownRange = 12;// inches
const double kAutoStraightSlowdownSpeed = .1;

const double kVelocityLoopTime = .020; // 20ms
const double kVelocityDeadzone = .05;
const double kVelocityMultiplier = .03;

const double kManualDeadzone = .05;

const double kFinishTimer = 0;
const double kSkipStepThreshold = .05;
const double kSkipStepTime = 1;

Drive::Drive(Lights *lights) {
    this->lights = lights;
    gyro.SetSensitivity(.007); // for ADW22307 Gyro (7 mV)
    gyro.InitGyro();

    motorLeftRear.SetInverted(true);
    motorLeftFront.SetInverted(true);

    gyroWarning.Start();
}

void Drive::reset() {
    encLeft.Reset();
    encRight.Reset();

    motorLeftFront.Set(0);

    motorLeftRear.Set(0);
    motorLeftRear.Set(0);
    motorRightFront.Set(0);
    motorRightRear.Set(0);

     for(int i = 0; i < kCircularHistorySize; i++) {
        circularHistoryLeft[i] = 0;
        circularHistoryRight[i] = 0;
    }

    velocityLeft = 0;
    velocityRight = 0;
}

void Drive::process() {
    double outputSpeedLeft = 0;
    double outputSpeedRight = 0;

    switch(driveState) {
        case STOPPED:
            if(!backdriveEncBroken) {
                if(fabs(velocityLeft) > kVelocityDeadzone) {
                        outputSpeedLeft = -velocityLeft * kVelocityMultiplier;
                }
                else {
                    outputSpeedLeft = 0;
                }

                if(fabs(velocityRight) > kVelocityDeadzone) {
                        outputSpeedRight = -velocityRight * kVelocityMultiplier;
                }
                else {
                    outputSpeedRight = 0;
                }
            }
            else {
                outputSpeedLeft = 0;
                outputSpeedRight = 0;
            }
          
        break;
        case AUTO_STRAIGHT:
        {
            double distanceToTargetL = fabs(autoDistance) - fabs((encLeft.Get() - autoStartEncL) / kTicksPerInch);
            double distanceToTargetR = fabs(autoDistance) - fabs((encRight.Get() - autoStartEncR) / kTicksPerInch);

            frc::SmartDashboard::PutNumber("drive_distanceL", distanceToTargetL);
            frc::SmartDashboard::PutNumber("drive_distanceR", distanceToTargetR);

            if(fabs(distanceToTargetL) > kAutoStraightSlowdownRange) { // if not close to target
                outputSpeedLeft = autoSpeed;
            }
            else { // if in range (should slow down)
                outputSpeedLeft = kAutoStraightSlowdownSpeed;
                if(distanceToTargetL < kAutoStraightAcceptableRange) {
                    outputSpeedLeft = 0;
                }
            }

            if(fabs(distanceToTargetR) > kAutoStraightSlowdownRange) { // if not close to target
                outputSpeedRight = autoSpeed;
            }
            else { // if in range (should slow down)
                outputSpeedRight = kAutoStraightSlowdownSpeed;
                if(distanceToTargetR < kAutoStraightAcceptableRange) {
                    outputSpeedRight = 0;
                }
            }            


            if(autoDistance < 0) {
                outputSpeedLeft *= -1;
                outputSpeedRight *= -1;
            }

            outputSpeedLeft = fclamp(outputSpeedLeft, -autoSpeed, autoSpeed);
            outputSpeedRight = fclamp(outputSpeedRight, -autoSpeed, autoSpeed);

            if(distanceToTargetL < kAutoStraightAcceptableRange && distanceToTargetR < kAutoStraightAcceptableRange) {
                autoFinished = true;
                finishTimer.Reset();
                finishTimer.Start();
                driveState = STOPPED;

                outputSpeedLeft = 0;
                outputSpeedRight = 0;
            }

            if(fabs((velocityLeft + velocityRight)/2) < kSkipStepThreshold) {
                // drive train isn't moving, probably up against a wall
                skippingAutoStep = true;
                skipTimer.Start();
            }
            else {
                skippingAutoStep = false;
            }
        }
        break;
        case AUTO_PIVOT:
        {
            double difference = autoAngle - gyro.GetAngle();

            // frc::SmartDashboard::PutNumber("drive_pivot_difference", difference);
            
            if(fabs(difference) < kAutoPivotAcceptableRange) { // in range, done
                autoFinished = true;
                finishTimer.Reset();
                finishTimer.Start();
                driveState = STOPPED;

                outputSpeedLeft = 0;
                outputSpeedRight = 0;
            }
            else { // not in range
                if(fabs(difference) > kAutoPivotSlowdownRange) { // fast range
                    outputSpeedRight = -autoSpeed;  
                    outputSpeedLeft = autoSpeed;
                }
                else { // slow range
                    outputSpeedRight = -kAutoPivotSlowdownSpeed;
                    outputSpeedLeft = kAutoPivotSlowdownSpeed;
                }

                if(difference < 0) {
                    outputSpeedRight *= -1;
                    outputSpeedLeft *= -1;
                }
                
            }
        }         
        break;
        case MANUAL:
            outputSpeedLeft = manualSpeedLeft;
            outputSpeedRight = manualSpeedRight;
        break;
    }

    outputSpeedLeft = fclamp(outputSpeedLeft, -1.0, 1.0);
    outputSpeedRight = fclamp(outputSpeedRight, -1.0, 1.0);

    motorLeftFront.Set(outputSpeedLeft);
    motorLeftRear.Set(outputSpeedLeft);
    
    motorRightFront.Set(outputSpeedRight);
    motorRightRear.Set(outputSpeedRight);

    shiftHistory(encLeft.Get(), encRight.Get());
    velocityLeft = (circularHistoryLeft[kCircularHistorySize-1]-circularHistoryLeft[0])/(kCircularHistorySize*kCircularHistorySize); // dPosition/dTime
    velocityRight = (circularHistoryRight[kCircularHistorySize-1]-circularHistoryRight[0])/(kCircularHistorySize*kCircularHistorySize); // dPosition/dTime

}

void Drive::setSpeed(double leftSpeed, double rightSpeed) {
    manualSpeedLeft = fclamp(leftSpeed, -1.0, 1.0);
    manualSpeedRight = fclamp(rightSpeed, -1.0, 1.0);

    if(fabs(manualSpeedLeft) < kManualDeadzone && fabs(manualSpeedRight) < kManualDeadzone) {
        driveState = STOPPED;
    }
    else {
        driveState = MANUAL;
    }
}

void Drive::autoForward(double distance, double speed) {
    driveState = AUTO_STRAIGHT;
    autoDistance = distance;
    autoSpeed = speed;
    finishTimer.Reset();
    finishTimer.Stop();
    autoFinished = false;

    skipTimer.Reset();
    skipTimer.Stop();

    autoStartEncL = encLeft.Get();
    autoStartEncR = encRight.Get();
}

void Drive::autoPivot(double angle, double speed) {
    driveState = AUTO_PIVOT;
    autoAngle = angle;
    autoSpeed = speed;

    finishTimer.Reset();
    finishTimer.Stop();
    autoFinished = false;

    gyro.Reset();
}

bool Drive::autoComplete() {
    bool finishedStep = autoFinished && finishTimer.Get() > kFinishTimer;
    bool skippingStep = skipTimer.Get() > kSkipStepTime;
    return finishedStep || skippingStep;
}

void Drive::debug(Feedback* feedback) {
    feedback->sendDouble("drive", "output left front", motorLeftFront.Get());
    feedback->sendDouble("drive", "output left rear", motorLeftRear.Get());
    feedback->sendDouble("drive", "output right front", motorRightFront.Get());
    feedback->sendDouble("drive", "output right rear", motorRightRear.Get());

    feedback->sendDouble("drive", "input left", manualSpeedLeft);
    feedback->sendDouble("drive", "input right", manualSpeedRight);

    feedback->sendDouble("drive", "encoder left", encLeft.Get());
    feedback->sendDouble("drive", "encoder right", encRight.Get());
    feedback->sendDouble("drive", "gyro", gyro.GetAngle());

    feedback->sendDouble("drive", "velocity left", velocityLeft / kTicksPerInch);
    feedback->sendDouble("drive", "velocity right", velocityRight / kTicksPerInch);

    feedback->sendBoolean("drive", "auto finished", autoFinished);
    feedback->sendDouble("drive", "auto timer", finishTimer.Get());
    feedback->sendBoolean("drive", "auto skipping", skippingAutoStep);
    feedback->sendDouble("drive", "auto skip timer", skipTimer.Get());

    feedback->sendBoolean("drive", "broken backdrive/enc", backdriveEncBroken);

    if(!gyroCalibrated) {
        if(gyroWarning.Get() > 1) {
            gyroWasRed = !gyroWasRed;
            gyroWarning.Reset();
        }
        feedback->sendDouble("thunderdashboard", "gyro", gyroWasRed);
    }
    else {
        feedback->sendDouble("thunderdashboard", "gyro", 0);
    }
}

void Drive::shiftHistory(double left, double right) {
     for(int i = 0; i < kCircularHistorySize-1; i++) {
        circularHistoryLeft[i] = circularHistoryLeft[i+1];
        circularHistoryRight[i] = circularHistoryRight[i+1];
    }
    circularHistoryLeft[kCircularHistorySize-1] = left;
    circularHistoryRight[kCircularHistorySize-1] = right;
}

void Drive::calibrateGyro() {
    if(gyroCalibrated == false) {
        gyroCalibrated = true;
        frc::SmartDashboard::PutNumber("thunderdashboard_gyro", 1);
        lights->setGyroCalibrating(true);
        gyro.Calibrate();
        lights->setGyroCalibrating(false);
        frc::SmartDashboard::PutNumber("thunderdashboard_gyro", 0);
    }
}

void Drive::setBackdriveBroken(bool broken) {
    backdriveEncBroken = broken;
}