#include "GamePiece.h"
#include <frc/smartdashboard/SmartDashboard.h>

const int kStartingCellCount = 3;
const double kShooterFlyWheelSpeed = 0.5;

// Time to wait for shot to complete when our shooter
// counter is broken. In seconds
const double kBrokenShotTimeout = 0.5;

const double kKeepIntakeAfterFive = 3;

const bool kRapidFire = true;


GamePiece::GamePiece(){
    intake = new Intake();
    helix = new Helix();
    shooter = new Shooter(kRapidFire);
    intakeState = NOTTAKE;
    shooterState = NOTSHOOTING;
    goGoGadgetShooter = false;
    shooterWantsPreheat = false;
    lastIntakeValue = intakeBeam.Get();
    shooterBeam.Reset();
    cellsShot = 0;
    intakeCounterBroken = false;
    transitionCounterBroken = false;
    shooterCounterBroken = false; 
    didWeHaveFiveCells = false;

    currentCellCount = kStartingCellCount;

}

void GamePiece::process(){
    enum {IN, OUT, NOT} HelixDir = NOT;
    int newShotCount = shooterBeam.Get();
    // ballWasShot has no meaning if shooterCounterBroken!
    bool ballWasShot = false;
    if (newShotCount != cellsShot){
        cellsShot = newShotCount;
        ballWasShot = true;
        if (currentCellCount != BALL_COUNT_UNKNOWN)
            // Don't decrement if our count was already unknown,
            // which also covers the weird case where we think we shot more
            // balls than we have (0 - 1 = -1 == BALL_COUNT_UNKNOWN)    
            currentCellCount--;
    }

    if (!intakeCounterBroken) {
        if(intakeBeam.Get() == true){
            lastIntakeValue = true;
        }
        if (intakeBeam.Get() == false && lastIntakeValue == true){
            // Do not change cell count away from UNKNOWN
            if (currentCellCount != BALL_COUNT_UNKNOWN){
                currentCellCount++;
            }
            lastIntakeValue = false;
        }
    }

    switch (intakeState){
        case INTAKE:
            if(currentCellCount >= 5){
                if(didWeHaveFiveCells == false){
                    keepIntakeAfterFive.Reset();
                    keepIntakeAfterFive.Start();
                }
                if(keepIntakeAfterFive.Get() >= kKeepIntakeAfterFive){
                    if (setNottake()){
                        HelixDir = NOT;
                    }
                    else{
                        HelixDir = IN;
                    }
                }
                else{
                    intake->setIntakePos(Intake::RETRACT);
                    intake->setIntakeDir(Intake::INTAKE);
                    HelixDir = IN;
                }
            }
            else{
                intake->setIntakePos(Intake::DEPLOY);
                intake->setIntakeDir(Intake::INTAKE);
                HelixDir = IN;
            }
            break;
        case OUTTAKE:
            intake->setIntakePos(Intake::RETRACT);
            intake->setIntakeDir(Intake::OUTTAKE);
            HelixDir = OUT;
            break;
        case NOTTAKE:
            if(currentCellCount == 0){
                HelixDir = NOT;
                intake->setIntakeDir(Intake::NOTTAKE);
                intake->setIntakePos(Intake::RETRACT);
            }
            else{
                if (setNottake()){
                    HelixDir = NOT;
                }
                else{
                    HelixDir = IN;
                }
            }
            break;
    }

    switch(shooterState){
        case NOTSHOOTING:
            if(goGoGadgetShooter == true){
                shooterState = WANTTOSHOOT;
            }
            shooter->setShooterSpinup(shooterWantsPreheat);
            break;
        case WANTTOSHOOT:
            if(currentCellCount != 0 && goGoGadgetShooter){
                shooter->setShooterSpinup(true);
                if (transitionCounterBroken) {
                    // Nothing really to do in this state
                    // as we can't tell when a ball is ready
                    // Just move on to shooting state
                    // with helix running
                    HelixDir = IN;
                    shooterState = SHOOT;
                    if (shooterCounterBroken) {
                        brokenShotTimer.Reset();
                        brokenShotTimer.Start();
                    }

                } else if (helixTopBeam.Get() == false){
                    shooterState = SHOOT;
                    HelixDir = NOT;
                    if (shooterCounterBroken) {
                        brokenShotTimer.Reset();
                        brokenShotTimer.Start();
                    }
                }
                else{
                    HelixDir = IN;
                }

                if(kRapidFire) {
                    shooterState = SHOOT;
                }
            }
            else{
                shooterState = NOTSHOOTING;
            }
            break;
        case SHOOT:
            shooter->setShooterSpinup(true);
            if(kRapidFire)
            {
                shooter->feedPowerCell(true);
                if(goGoGadgetShooter == false) {
                    shooterState = NOTSHOOTING;
                    shooter->feedPowerCell(false);
                    shooter->setShooterSpinup(false);
                }
                HelixDir = IN;   
            }
            else {
                if (transitionCounterBroken) {
                    HelixDir = IN;
                } else {
                    HelixDir = NOT;
                }

                if((shooterCounterBroken && brokenShotTimer.Get() < kBrokenShotTimeout) ||
                (!shooterCounterBroken && ballWasShot == false)){
                    shooter->feedPowerCell(true);
                }
                else{
                    shooter->feedPowerCell(false);
                    if(goGoGadgetShooter == true){
                        shooterState = WANTTOSHOOT;
                    }
                    else{
                        shooterState = NOTSHOOTING;
                    }
                }

            }
            break;

    }
    if(HelixDir == IN){
        helix->SetOn();
    }
    else if(HelixDir == OUT){
        helix->SetReverse();
    }
    else{
        helix->SetOff();
    }

    // Tell subsystems to process
    intake->process();
    helix->Process();
    shooter->process();

    ballWasShotLast = ballWasShot;

    if(currentCellCount >= 5){
        didWeHaveFiveCells = true;
    }
    else{
        didWeHaveFiveCells = false;
    }
}

void GamePiece::reset(int ballCount = -1){
    if(ballCount >= 0 && !intakeCounterBroken && !shooterCounterBroken){  
        currentCellCount = ballCount;
    }
    /* else if(ballCount == -1) {
        currentCellCount = frc::SmartDashboard::GetNumber("thunderdashboard_starting_count", 0);
    } */
    intakeState = NOTTAKE;
    shooterState = NOTSHOOTING;
    goGoGadgetShooter = false;
    shooterWantsPreheat = false;
    shooter->reset();
    helix->SetOff();
    intake->reset();
    
    cellsShot = shooterBeam.Get();
    
}

void GamePiece::debug(Feedback* feedback){
    feedback->sendDouble("thunderdashboard", "ballcount", currentCellCount);

    switch(intakeState){
        case INTAKE:
            feedback->sendString("gamepiece", "intakeState", "INTAKE");
            break;
        case OUTTAKE:
            feedback->sendString("gamepiece", "intakeState", "OUTTAKE");
            break;
        case NOTTAKE:
            feedback->sendString("gamepiece", "intakeState", "NOTTAKE");
            break;
    }

    switch(shooterState){
        case NOTSHOOTING:
            feedback->sendString("gamepiece", "shooterState" , "NOTSHOOTING");
            break;
        case WANTTOSHOOT:
            feedback->sendString("gamepiece" , "shooterState" , "WANTTOSHOOT");
            break;
        case SHOOT:
            feedback->sendString("gamepiece" , "shooterState" , "SHOOT");
            break;
    }

    feedback->sendDouble("gamepiece" , "shotCount" , cellsShot);
    feedback->sendDouble("gamepiece" , "currentCellCount" , currentCellCount);
    feedback->sendBoolean("gamepiece" , "intakeBeam" , intakeBeam.Get());
    feedback->sendBoolean("gamepiece" , "helixTopBeam" , helixTopBeam.Get());
    feedback->sendDouble("gamepiece" , "shooterSensorCount" , shooterBeam.Get());
    feedback->sendBoolean("gamepiece", "brokenIntakeCounter", intakeCounterBroken);
    feedback->sendBoolean("gamepiece", "brokenTransitionCounter", transitionCounterBroken);
    feedback->sendBoolean("gamepiece", "brokenShooterCounter", shooterCounterBroken);
    feedback->sendBoolean("gamepiece", "ball was shot", ballWasShotLast);


    feedback->sendBoolean("a_syscheck", "intake_beam_dio0", intakeBeam.Get());
    feedback->sendBoolean("a_syscheck", "helix_top_beam_dio2", helixTopBeam.Get());
    feedback->sendDouble("a_syscheck", "shooter_beam_dio3", shooterBeam.Get());

    intake->debug(feedback);
    helix->Debug(feedback);
    shooter->debug(feedback);
}

void GamePiece::setIntakeState(IntakeState intState){
    intakeState = intState;
}

void GamePiece::activateShooter(bool shooterActivate){
    goGoGadgetShooter = shooterActivate;
}

// Returns true if helix does not need to run in the intake not-take state.
// Returns false if helix should run in that state.
bool GamePiece::setNottake(){
    intake->setIntakePos(Intake::RETRACT);
    intake->setIntakeDir(Intake::NOTTAKE);
            
    if(transitionCounterBroken || helixTopBeam.Get() == false){
        return true;
    }
    else{
        return false;
    }
}

void GamePiece::setShooterPreheat(bool preheat){
    shooterWantsPreheat = preheat;
}
void GamePiece::hood(Shooter::HoodPosition position){
    shooter->hood(position);
}
void GamePiece::setShooterSpeed(double speed){
    shooter->setShooterSpeed(speed);
}
void GamePiece::pivotSensorBroken(bool broken){
    intake->setPivotSensorBroken(broken);
}
void GamePiece::setIntakeCounterBroken(bool broken){
    intakeCounterBroken = broken;
    if (broken)
        // Once broken, we have no idea what our count is
        currentCellCount = BALL_COUNT_UNKNOWN;
}
void GamePiece::setTransitionCounterBroken(bool broken){
    transitionCounterBroken = broken;
}
void GamePiece::setShooterBallCounterBroken(bool broken){
    if (broken && !shooterCounterBroken) {
        // Once broken, we have no idea what our count is
        currentCellCount = BALL_COUNT_UNKNOWN;
        // We just transitioned to broken from not
        // Make sure to set up the shot timer in the very very unlikely
        // chance we were set to broken while waiting for a shot to happen
        brokenShotTimer.Reset();
        brokenShotTimer.Start();
    }
    shooterCounterBroken = broken;
}
int GamePiece::getCellCount(){
    return currentCellCount;
}

void GamePiece::lights(Lights* lights) {
    lights->setBallCount(getCellCount());
}

void GamePiece::setIntakeIdleMode(bool enabled){
    intake->setIntakeIdleMode(enabled);
}

void GamePiece::handleLog(Log* log) {
    log->setBallCount(currentCellCount);
    log->setHelixSensor(helixTopBeam.Get());
    log->setIntakeSensor(intakeBeam.Get());
    log->setShooterCounter(shooterBeam.Get());
    log->setShooterState(shooterState);
    log->setIntakeState(intakeState);

    intake->handleLog(log);

}