#include "Hang.h"

// Height Thresholds, using neo shaft rotations
//UP
const double kMaxHeight = 560; // will not go up above this height 
const double kFastUpHeight = 460; // will drive up faster for range that will not be stopped in

//DOWN
const double kMaxRetractHeight = 0; // will not go down below this height
const double kParkHeight = 50; // goes extremely slowly below this height
const double kFastDownHeight = 100; // when moving down, it will get faster until this value, at which point the robot should be off the ground
const double kPrecisionDownHeight = 460; // when moving down, it will initially move slowly until this limit is reached

const double kExtendSpeed = .3; // fine precision up speed
const double kRetractSpeed = .3; // fine precision down speed

const double kDisengageThreshold = 2; // 22 degrees / 360 degrees = .06 rotations (10)
const double kDisengageSpeed = .2; // absolute

const double kFastUpSpeed = 1; // fast up speed .8
const double kNormalUpSpeed = .5; // normal up speeed .5

const double kPrecisionDownSpeed = .2;
const double kFastDownSpeed = 1; // fast down speed until off the ground 1
const double kParkDownSpeed = .1;

const double kServoEngaged = .3; // .75
const double kServoDisengaged = .72; // .2


Hang::Hang() : winch(ThunderSparkMax::create(ThunderSparkMax::MotorID::Hang)) {
    setRatchetEngaged(true);
    winch->SetEncoder(0);
    winch->SetInverted(true);
    winch->SetIdleMode(ThunderSparkMax::BRAKE);
    winch->SetOpenLoopRampRate(.5);
}

void Hang::process() {
    switch(currentState) {
        case STOPPED:
            winch->Set(0);
            setRatchetEngaged(true);

            if(moveDirection == UP && winch->GetEncoder() < kMaxHeight) {
                currentState = WINCH_DISENGAGE;
                preEngageEncoder = winch->GetEncoder();
            }
            
            if(moveDirection == DOWN && winch->GetEncoder() > kMaxRetractHeight) {
                currentState = MOVING_DOWN;
            }
        break;
        case WINCH_DISENGAGE:
            setRatchetEngaged(false);
            if(winch->GetEncoder() > preEngageEncoder-kDisengageThreshold) {
                winch->Set(-fabs(kDisengageSpeed));
            }
            else if(winch->GetEncoder() <= preEngageEncoder-kDisengageThreshold) {
                winch->Set(0);
                currentState = MOVING_UP;
            }
        break;
        case MOVING_UP:
            setRatchetEngaged(false);
            if(winch->GetEncoder() < kFastUpHeight) {
                winch->Set(kFastUpSpeed);
            }
            else if(winch->GetEncoder() < kMaxHeight) {
                winch->Set(kNormalUpSpeed);
            }
            else {
                winch->Set(0);
                currentState = STOPPED;
            }

            if(moveDirection != UP) {
                currentState = STOPPED;
            }
        break;
        case MOVING_DOWN:
            if(winch->GetEncoder() > kFastDownHeight) {
                if(!slowRetract) {
                    winch->Set(-kFastDownSpeed);
                }
                else {
                    winch->Set(-kPrecisionDownSpeed);
                }
            }
            else if(winch->GetEncoder() > kMaxRetractHeight) {
                if(winch->GetEncoder() < kParkHeight) {
                    winch->Set(-kParkDownSpeed);
                }
                else {
                    winch->Set(-kRetractSpeed);
                }
            }        
            
            if(winch->GetEncoder() < kMaxRetractHeight) {
                currentState = STOPPED;
            }

            if(moveDirection != DOWN) {
                currentState = STOPPED;
            }            

        break;
    }

    if(!zeroSensorBroke) {
        if(lastSensorReading == true && zeroSensor.Get() == false) {
            winch->SetEncoder(0);
        }
        lastSensorReading = zeroSensor.Get();
    }
}

void Hang::reset() {



}

void Hang::move(Hang::HangMovement position) {
    moveDirection = position;
}

void Hang::setRatchetEngaged(bool engaged) {
    ratchetEngaged = engaged;
    if(engaged) {
        ratchet.Set(kServoEngaged);
    }
    else {
        ratchet.Set(kServoDisengaged);
    }
}

void Hang::debug(Feedback* feedback) {

    feedback->sendDouble("thunderdashboard", "hang_pos", 100 * (winch->GetEncoder() / kMaxHeight));

    feedback->sendDouble("hang", "speed", winch->Get());
    feedback->sendDouble("hang", "encoder", winch->GetEncoder());
    feedback->sendDouble("hang", "ratchet", ratchetEngaged);
    feedback->sendString("hang", "zero sensor", zeroSensor.Get() ? "true" : "false");

    const char* stateName = "";
    switch(currentState) {
        case STOPPED:
            stateName = "STOPPED";
        break;
        case WINCH_DISENGAGE:
            stateName = "WINCH_DISENGAGE";
        break;
        case MOVING_UP:
            stateName = "MOVING_UP";
        break;
        case MOVING_DOWN:
            stateName = "MOVING_DOWN";
        break;
    }
    feedback->sendString("hang", "state", stateName);

    const char* directionName = "";
    switch(moveDirection) {
        case UP:
            directionName = "UP";
        break;
        case STOP:
            directionName = "STOP";
        break;
        case DOWN:
            directionName = "DOWN";
        break;
    }
    feedback->sendString("hang", "direction", directionName);
    feedback->sendBoolean("hang", "sensor broken", zeroSensorBroke);
}
void Hang::zeroSensorBroken(bool broke){
    zeroSensorBroke = broke;
}

void Hang::enableSlowRetract(bool enable) {
    slowRetract = enable;
}

void Hang::setHangIdleMode(bool enabled) {
        if(enabled){
        winch->SetIdleMode(ThunderSparkMax::BRAKE);
    }
    else{
        winch->SetIdleMode(ThunderSparkMax::COAST);
    }
}