#include "Intake.h"

const double kIntakeSpinSpeed = .80;
const double kOutakeSpinSpeed = .45;

const double kIntakeDeploySpeed = .8; //positive to run the motor forwards

const double kIntakePivotFast = -1; //negative to run the motor backwards
const double kIntakePivotMed = -0.8; //negative to run the motor backwards
const double kIntakePivotSlow = -0.4; //negative to run the motor backwards

const double kIntakeBottom = 45; //at bottom 54
const double kIntakeTwoThirdRange = 36;
const double kIntakeOneThirdRange = 18; //bottom is 66.45
const double kIntakeTop = 2;

//const double kIntakePivotEncValue = 90;           Is this used?

//for reference: NEO 500 Encoder 42 counts per rev.

Intake::Intake() :
    intakeDeployMotor(ThunderSparkMax::create(ThunderSparkMax::MotorID::IntakePivot)),
    intakeSpinMotor(ThunderSparkMax::create(ThunderSparkMax::MotorID::IntakeBeaterBars))
{
    intakeSpinMotor->SetIdleMode(ThunderSparkMax::COAST); //sets motor to brake mode
    intakeDeployMotor->SetOpenLoopRampRate(0.25); //sets the time to full speed to .25s
    intakeSpinMotor->SetOpenLoopRampRate(0.5); //sets the time to full speed to .5s

    intakePivBroken = false;
    intakeDeployMotor->SetEncoder(0);
    intakeDir = NOTTAKE;
    intakePos = RETRACT;
}

void Intake::process() {
    if(intakePivBroken == false && encZeroSensor.Get() == false){
        intakeDeployMotor->SetEncoder(0);
    }

    switch(intakeDir) {
        case INTAKE:
            setIntakeSpinMotorSpeed(kIntakeSpinSpeed); 
            break;
        case OUTTAKE:
            setIntakeSpinMotorSpeed(-kOutakeSpinSpeed);
            break;
        case NOTTAKE:
            setIntakeSpinMotorSpeed(0);
            break;
    }

    switch(intakePos) {
        case DEPLOY:
            if(startup == false){
                if(intakeDeployMotor->GetEncoder() < kIntakeBottom){
                    setIntakeDeployMotorSpeed(kIntakeDeploySpeed);
                }
                else {
                    setIntakeDeployMotorSpeed(0);
                }
            }
            else{
                if(encZeroSensor.Get() == true){
                    setIntakeDeployMotorSpeed(kIntakePivotSlow);
                }
                else{
                    setIntakeDeployMotorSpeed(0);
                    startup = false;
                    intakeDeployMotor->SetEncoder(0);
                }
            }
                break;
        case RETRACT:
            if(startup == false){
                if(intakeDeployMotor->GetEncoder() >= kIntakeBottom){
                    setIntakeDeployMotorSpeed(kIntakePivotFast);
                }
                if (intakeDeployMotor->GetEncoder() <= kIntakeBottom && intakeDeployMotor->GetEncoder() >= kIntakeTwoThirdRange){
                    setIntakeDeployMotorSpeed(kIntakePivotMed);
                }
                if(intakeDeployMotor->GetEncoder() <= kIntakeTwoThirdRange && intakeDeployMotor->GetEncoder() >= kIntakeOneThirdRange){
                    setIntakeDeployMotorSpeed(kIntakePivotSlow);
                }
                if(intakeDeployMotor->GetEncoder() <= kIntakeOneThirdRange && intakeDeployMotor->GetEncoder() >= kIntakeTop){
                    setIntakeDeployMotorSpeed(kIntakePivotSlow);
                }
                if(intakeDeployMotor->GetEncoder() <= kIntakeTop){
                    setIntakeDeployMotorSpeed(0);
                }
            }
            else{
                if(encZeroSensor.Get() == true){
                    setIntakeDeployMotorSpeed(kIntakePivotSlow);
                }
                else{
                    setIntakeDeployMotorSpeed(0);
                    startup = false;
                    intakeDeployMotor->SetEncoder(0);
                }
            }
            break;
    }
}

void Intake::reset(){
    intakePos = RETRACT;
    intakeDir = NOTTAKE;
    
    if(intakePivBroken == false){
        startup = true;
    }
}

void Intake::debug(Feedback* feedback) {
 feedback->sendDouble("thunderdashboard", "intake_pos",  intakeDeployMotor->GetEncoder() >= kIntakeBottom ? 0 : 1);


    switch(intakeDir){
        case INTAKE:
            feedback->sendString("intake", "intakeDirection", "IN");
            break;
        case OUTTAKE:
            feedback->sendString("intake", "intakeDirection", "OUT");
            break;
        case NOTTAKE:
            feedback->sendString("intake", "intakeDirection", "NOT");
            break;
    }

    switch(intakePos){
        case DEPLOY:
            feedback->sendString("intake", "intakePosition", "DEPLOYED");
            break;
        case RETRACT:
            feedback->sendString("intake", "intakePosition", "RETRACTED");
            break;
    }

    feedback->sendDouble("intake", "intakeDeployEncValue", intakeDeployMotor->GetEncoder());
    feedback->sendBoolean("intake", "encZeroSensor", encZeroSensor.Get());
    feedback->sendBoolean("intake", "intakePivBroken", intakePivBroken);

    feedback->sendBoolean("a_syscheck", "intake_zero_dio8", encZeroSensor.Get());
    feedback->sendDouble("intake", "spin current", intakeSpinMotor->getSparkCurrent());
}

void Intake::setIntakeDir(IntakeDir direction){
    intakeDir = direction;
}

void Intake::setIntakePos(IntakePos position){
    intakePos = position;
} 

void Intake::setIntakeSpinMotorSpeed(double speed){
    intakeSpinMotor->Set(speed);
}

void Intake::setIntakeDeployMotorSpeed(double speed){
    intakeDeployMotor->Set(speed); 
}

void Intake::zeroPivotEncoder(){
    if(intakePivBroken == false){
        if(encZeroSensor.Get() == false){
            intakeDeployMotor->SetEncoder(0);
        }
    }
}

void Intake::setPivotSensorBroken(bool broken){
    intakePivBroken = broken;
}

void Intake::setIntakeIdleMode(bool enabled){
    if(enabled){
        intakeDeployMotor->SetIdleMode(ThunderSparkMax::BRAKE);
    }
    else{
        intakeDeployMotor->SetIdleMode(ThunderSparkMax::COAST);
    }
}

void Intake::handleLog(Log* log) {
    log->setIntakeCurrent(intakeSpinMotor->getSparkCurrent());
}