#include "Shooter.h"
#include "ShooterPID.h"

const double kServoWall = .47;
const double kServoTargetCorner = .44;
const double kServoLine = .35;
const double kServoTrench = .35;

const double kVelocityScaleFactor = .5; // 2:1 gear ratio

const double kReadyToShootDeadzone = .2;

const double kPollCooldown = .05;
const double kHistoryThreshold = .5;
const double kVelocityThreshold = .5;
const double kMovingDeadzone = 100;

const double kTransitionFeedSpeed = 1;

const double kAverageTolerance = 200; // 100 rpm tolerance

const int kRPMWall = 3250; // 2950
const int kRPMTargetCorner = 3200;
const int kRPMLine = 4300;
const int kRPMTrench = 4800; // 80%

const int kNeo550MaxRPM = 916;

Shooter::Shooter(bool rapidFire) : primerMotor(ThunderSparkMax::create(ThunderSparkMax::MotorID::ShooterPrimer)),
                     shooterMotorMaster(ThunderSparkMax::create(ThunderSparkMax::MotorID::ShooterRight)),
                     shooterMotorSlave(ThunderSparkMax::create(ThunderSparkMax::MotorID::ShooterLeft)),
                     transitionMotor(ThunderSparkMax::create(ThunderSparkMax::MotorID::StorageTransitionToShooter)),
                     pidController(shooterMotorMaster->getSparkPID()),
                     masterShooterEnc(shooterMotorMaster->getSparkEncoder())
{
    shooterMotorMaster->SetIdleMode(ThunderSparkMax::IdleMode::COAST);

    shooterMotorMaster->SetInverted(true);

    shooterMotorSlave->SetIdleMode(ThunderSparkMax::IdleMode::COAST);
    shooterMotorSlave->Follow(shooterMotorMaster, true);

    cooldown.Start();

    rapidFiring = rapidFire;

    primerMotor->SetInverted(true);

    for(int i = 0; i < kHistoryCount; i++) {
        velocityHistory[i] = 0;
    }

    pidController->SetP(ShooterPID::kP);
    pidController->SetI(ShooterPID::kI);
    pidController->SetD(ShooterPID::kD);
    pidController->SetFF(ShooterPID::kF);
    pidController->SetIZone(ShooterPID::kIZone);
    pidController->SetOutputRange(ShooterPID::kMinOutput, ShooterPID::kMaxOutput);
}

void Shooter::reset()
{
    primerMotor->Set(0);
    shooterMotorMaster->Set(0);
    transitionMotor->Set(0);
    pidController->SetReference(0, rev::ControlType::kVelocity);
}

void Shooter::debug(Feedback *feedback)
{
    feedback->sendDouble("shooter", "velocity", masterShooterEnc->GetVelocity());
    feedback->sendDouble("shooter", "output master", shooterMotorMaster->Get());
    feedback->sendDouble("shooter", "output slave", shooterMotorSlave->Get());
    feedback->sendBoolean("shooter", "want to spin", wantToSpin);
    feedback->sendBoolean("shooter", "want to shoot", wantToShoot);

    const char* presetName = "";
    switch(hoodPos) {
        case PANEL:
        presetName = "panel";
        break;
        case LINE:
        presetName = "line";
        break;
        case PORT:
        presetName = "port";
        break;
        case MANUAL:
        presetName = "manual";
        break;
        case TARGET_CORNER:
        presetName = "corner of targetzone";
        break;
    }
    feedback->sendString("shooter", "preset", presetName);
    feedback->sendDouble("shooter", "hood position", hoodServo.Get());

    feedback->sendDouble("shooter", "set point", setPoint);

    if(manualControl) {
        feedback->sendDouble("thunderdashboard", "shooter_speed", 100*desiredSpeed);
    }
    else {
        feedback->sendDouble("thunderdashboard", "shooter_speed", 100*(masterShooterEnc->GetVelocity() / ShooterPID::kMaxRPM));
    }

    feedback->sendDouble("thunderdashboard", "shooter_hood", hoodServo.Get() * 100);

    feedback->sendDouble("shooter", "manual speed", desiredSpeed);
    feedback->sendBoolean("shooter", "manual enabled", manualControl);

    if(initDebug) {
        initDebug = false;

        feedback->sendEditableDouble("shooter", "pid f", ShooterPID::kF);
        feedback->sendEditableDouble("shooter", "pid p", ShooterPID::kP);
    }
    else {
        if(feedback->getEditableDouble("shooter", "pid f", ShooterPID::kF) != ShooterPID::kF) {
            pidController->SetFF(feedback->getEditableDouble("shooter", "pid f", ShooterPID::kF));
        }

        if(feedback->getEditableDouble("shooter", "pid p", ShooterPID::kP) != ShooterPID::kP) {
            pidController->SetP(feedback->getEditableDouble("shooter", "pid p", ShooterPID::kP));
        }
    }

    feedback->sendDouble("shooter", "primer output", primerMotor->getSparkEncoder()->GetVelocity() / kNeo550MaxRPM);


    //dashboard graphing

    feedback->sendString("shooter", "chart_plotmin", "0"); // min
    feedback->sendString("shooter", "chart_plotmax", "0"); // max
    feedback->sendString("shooter", "chart_platlab", "RPM"); // label
    feedback->sendString("shooter", "chart_plotctr", "0"); // centerline

    feedback->sendString("shooter", "chart_plot", "%d", masterShooterEnc->GetVelocity());


}
void Shooter::process()
{
    double finalTransitionSpeed = 0;
    double finalPrimerSpeed = 0;
    double shooterSetPoint = 0;

    if(wantToSpin) {
            switch(hoodPos) {
                case PANEL:
                shooterSetPoint = kRPMTrench;
                break;
                case LINE:
                shooterSetPoint = kRPMLine;
                break;
                case PORT:
                shooterSetPoint = kRPMWall;
                break;
                case TARGET_CORNER:
                shooterSetPoint = kRPMTargetCorner;
                break;
                case MANUAL:
                shooterSetPoint = ShooterPID::kMaxRPM;
                break;
            }

        finalPrimerSpeed = 1;
        if(wantToShoot && isAtVelocity()) {
            finalTransitionSpeed = kTransitionFeedSpeed;
        }
    }
    else {
        
    }


    transitionMotor->Set(finalTransitionSpeed);
    primerMotor->Set(finalPrimerSpeed);

    if(manualControl && wantToSpin) {
        pidController->SetReference(desiredSpeed * ShooterPID::kMaxRPM, rev::ControlType::kVelocity);
    }
    else {
        pidController->SetReference(shooterSetPoint, rev::ControlType::kVelocity);
    }
    setPoint = shooterSetPoint;
    

    if(cooldown.Get() > kPollCooldown) {
        cooldown.Reset();
        shiftVelocity(masterShooterEnc->GetVelocity());
    }

    switch(hoodPos) {
        case LINE:
            hoodServo.Set(kServoLine);
            break;
        case PANEL:
            hoodServo.Set(kServoTrench);
            break;
        case PORT:
            hoodServo.Set(kServoWall);
            break;
        case TARGET_CORNER:
            hoodServo.Set(kServoTargetCorner);
            break;
        case MANUAL:
            hoodServo.Set(hoodPosManual);
            break;
    }
}

void Shooter::setShooterSpinup(bool shouldSpin)
{
    if(wantToSpin == false && shouldSpin == true) {
        for(int i = 0; i < kHistoryCount; i++) {
            shiftVelocity(0);
        }
    }
    wantToSpin = shouldSpin;
    
}
void Shooter::feedPowerCell(bool feed)
{
    wantToShoot = feed;
}

void Shooter::setShooterSpeed(double speed)
{
    desiredSpeed = fabs(speed);
}


void Shooter::hood(HoodPosition position)
{
    hoodPos = position;
}

bool Shooter::isAtVelocity() {
    if(rapidFiring) {
        return true;
    }
    else {
        return masterShooterEnc->GetVelocity() > setPoint - kAverageTolerance;
    }
}

void Shooter::shiftVelocity(double velocity) {
    for(int i = 0; i < kHistoryCount-1; i++) {
        velocityHistory[i] = velocityHistory[i+1];
    }
    velocityHistory[kHistoryCount-1] = velocity;
}

void Shooter::enableManualControl(bool enabled) {
    manualControl = enabled;
}

void Shooter::hoodManual(double position) {
    hoodPos = MANUAL;
    hoodPosManual = position;
}