#include "ThunderSparkMax.h"
#include "IOMap.h"

class ThunderSparkMaxImpl : public ThunderSparkMax {
    public:
        ThunderSparkMaxImpl(const char *implName, int id);
        virtual rev::CANSparkMax *getSparkMax();
        virtual void Set(double speed);
        virtual double Get();
        
        virtual void SetOpenLoopRampRate(double rate);

        // Returns rotations of encoder
        virtual double GetEncoder() ;
        virtual void SetEncoder(double rotations);

        virtual void SetIdleMode(IdleMode idleMode);
        virtual void SetInverted(bool inverted);
        virtual void Follow(ThunderSparkMax *leader, bool invertOutput = false);

        virtual rev::CANPIDController* getSparkPID();
        virtual rev::CANEncoder* getSparkEncoder();
        virtual double getSparkCurrent();

    protected:
        const char *implName;
        int id;

        void printWarning(bool *hasPrinted, const char *description);
        double lastSpeed;
        bool printedPIDWarning;
        bool printedEncoderWarning;
        bool printedIdleWarning;
        bool printedInvertWarning;
        bool printedFollowWarning;
        bool printedRampWarning;
        bool printedCurrentWarning;
};

ThunderSparkMaxImpl::ThunderSparkMaxImpl(const char *implName, int id) :
        implName(implName),
        id(id),
        lastSpeed(0),
        printedPIDWarning(false),
        printedEncoderWarning(false),
        printedIdleWarning(false),
        printedInvertWarning(false),
        printedFollowWarning(false),
        printedRampWarning(false),
        printedCurrentWarning(false)
{
    
}

rev::CANSparkMax *ThunderSparkMaxImpl::getSparkMax()
{
    return NULL;
}

void ThunderSparkMaxImpl::Set(double speed)
{
    lastSpeed = speed;
}

double ThunderSparkMaxImpl::Get()
{
    return lastSpeed;
}

void ThunderSparkMaxImpl::SetOpenLoopRampRate(double rate)
{
    printWarning(&printedRampWarning, "Output Ramping");
}
        
double ThunderSparkMaxImpl::GetEncoder()
{
    printWarning(&printedEncoderWarning, "Encoder");
    return 0;
}

void ThunderSparkMaxImpl::SetEncoder(double rotations)
{
    printWarning(&printedEncoderWarning, "Encoder");
}

void ThunderSparkMaxImpl::SetIdleMode(IdleMode idleMode)
{
    printWarning(&printedIdleWarning, "Idle Mode");
}

void ThunderSparkMaxImpl::SetInverted(bool inverted)
{
    printWarning(&printedInvertWarning, "Invert of output");
}
void ThunderSparkMaxImpl::Follow(ThunderSparkMax *leader, bool invertOutput)
{
    printWarning(&printedFollowWarning, "Follower Mode");
}

rev::CANPIDController* ThunderSparkMaxImpl::getSparkPID() {
    printWarning(&printedPIDWarning, "Native PID");
    return NULL;
}

rev::CANEncoder* ThunderSparkMaxImpl::getSparkEncoder() {
    printWarning(&printedEncoderWarning, "SPARK MAX Encoder");
    return NULL;
}

double ThunderSparkMaxImpl::getSparkCurrent()
{
    printWarning(&printedCurrentWarning, "Current");
    return 0;
}

void ThunderSparkMaxImpl::printWarning(bool *hasPrinted, const char *description)
{
    if (!(*hasPrinted)) {
        printf("WARNING: %s %d does not support %s!\n", implName, id, description);
        *hasPrinted = true;
    }
}







#include <frc/Jaguar.h>
class ThunderSMJagImpl : public ThunderSparkMaxImpl {
    public:
        ThunderSMJagImpl(int pwmChannel);
        virtual void Set(double speed);

    private:
        frc::Jaguar jaguar;
};

#include <ctre/phoenix.h>
class ThunderSMTalonImpl : public ThunderSparkMaxImpl {
    public:
        ThunderSMTalonImpl(int canId);
        virtual void Set(double speed);

        virtual void SetOpenLoopRampRate(double rate);

        virtual double GetEncoder();
        virtual void SetEncoder(double rotations);
        virtual void SetInverted(bool invert);

        virtual void SetIdleMode(IdleMode idleMode);
    private:
        TalonSRX talon;
};

class ThunderSMCANImpl : public ThunderSparkMaxImpl {
    public:
        ThunderSMCANImpl(int canId);
        virtual void Set(double speed);
        
        virtual void SetOpenLoopRampRate(double rate);

        virtual double GetEncoder();
        virtual double Get();
        virtual void SetEncoder(double rotations);
        virtual void SetInverted(bool invert);

        virtual void SetIdleMode(IdleMode idleMode);
        virtual void Follow(ThunderSparkMax *leader, bool invertOutput = false);

        virtual rev::CANSparkMax *getSparkMax() { return &spark; }
        virtual rev::CANPIDController *getSparkPID() { return &pid; }
        virtual rev::CANEncoder* getSparkEncoder() { return &enc; }

        virtual double getSparkCurrent() { return spark.GetOutputCurrent();}

    private:
        rev::CANSparkMax spark;
        rev::CANEncoder enc;
        rev::CANPIDController pid;
};

#include <frc/PWMSparkMax.h>
class ThunderSMPWMImpl : public ThunderSparkMaxImpl {
    public:
        ThunderSMPWMImpl(int canId);
        virtual void Set(double speed);
    private:
        frc::PWMSparkMax spark;
};


struct MotorConfig {
    enum ControllerType {
        Jaguar,
        Talon,
        PWMSparkMax,
        CANSparkMax
    } controllerType;

    int channelOrId;

    MotorConfig(ControllerType controllerType, int channelOrId) : controllerType(controllerType), channelOrId(channelOrId) {}
};




ThunderSparkMax::ThunderSparkMax()
{
}

ThunderSparkMax::~ThunderSparkMax()
{
}

ThunderSparkMax *ThunderSparkMax::create(MotorID id)
{
    static MotorConfig config[] = {
        MotorConfig(MotorConfig::ControllerType::PWMSparkMax, PWM_DRIVE_LF),    // DriveFrontLeft
        MotorConfig(MotorConfig::ControllerType::PWMSparkMax, PWM_DRIVE_RF),    // DriveFrontRight,
        MotorConfig(MotorConfig::ControllerType::PWMSparkMax, PWM_DRIVE_LR),    // DriveRearLeft,
        MotorConfig(MotorConfig::ControllerType::PWMSparkMax, PWM_DRIVE_RR),    // DriveRearRight,
#ifndef TEST_BOARD
        MotorConfig(MotorConfig::ControllerType::CANSparkMax, CAN_INTAKE_PIVOT),   // IntakePivot,
        MotorConfig(MotorConfig::ControllerType::CANSparkMax, CAN_INTAKE_BEATERBARS),   // IntakeBeaterBars,
        MotorConfig(MotorConfig::ControllerType::CANSparkMax, CAN_HELIX),    // StorageAgitator,
        MotorConfig(MotorConfig::ControllerType::CANSparkMax, CAN_STORAGE_SHOOT_TRANSITION),   // StorageTransitionToShooter,
        MotorConfig(MotorConfig::ControllerType::CANSparkMax, CAN_SHOOT_PRIMER),   // ShooterPrimer,
        MotorConfig(MotorConfig::ControllerType::CANSparkMax, CAN_SHOOT_LEFT),   // ShooterLeft,
        MotorConfig(MotorConfig::ControllerType::CANSparkMax, CAN_SHOOT_RIGHT),   // ShooterRight,
        MotorConfig(MotorConfig::ControllerType::CANSparkMax, CAN_CONTROL_SPINNER),   // WheelOfFortune,
        MotorConfig(MotorConfig::ControllerType::CANSparkMax, CAN_HANG_WINCH),   // Hang,
#else
        MotorConfig(MotorConfig::ControllerType::CANSparkMax, 10),   // IntakePivot,
        MotorConfig(MotorConfig::ControllerType::Talon, 8),   // IntakeBeaterBars,
        MotorConfig(MotorConfig::ControllerType::Jaguar, 2),    // StorageAgitator,
        MotorConfig(MotorConfig::ControllerType::Talon, 2),   // StorageTransitionToShooter,
        MotorConfig(MotorConfig::ControllerType::Jaguar, 3),   // ShooterPrimer,
        MotorConfig(MotorConfig::ControllerType::CANSparkMax, 9),   // ShooterLeft,
        MotorConfig(MotorConfig::ControllerType::Talon, 3),   // ShooterRight,
        MotorConfig(MotorConfig::ControllerType::Talon, 6),   // WheelOfFortune,
        MotorConfig(MotorConfig::ControllerType::Talon, 5),   // Hang,
#endif
    };

    ThunderSparkMax *c = NULL;

    switch (config[id].controllerType)
    {
        case MotorConfig::ControllerType::PWMSparkMax:
            c = new ThunderSMPWMImpl(config[id].channelOrId);
            break;
        case MotorConfig::ControllerType::CANSparkMax:
            c = new ThunderSMCANImpl(config[id].channelOrId);
            break;
        case MotorConfig::ControllerType::Talon:
            c = new ThunderSMTalonImpl(config[id].channelOrId);
            break;
        case MotorConfig::ControllerType::Jaguar:
            c = new ThunderSMJagImpl(config[id].channelOrId);
            break;
    }

    return c;
}



/**************************************************************/
// ThunderSMJagImpl

ThunderSMJagImpl::ThunderSMJagImpl(int pwmChannel) : ThunderSparkMaxImpl("Jaguar", pwmChannel), jaguar(pwmChannel)
{
}

void ThunderSMJagImpl::Set(double speed)
{
    jaguar.Set(speed);
    ThunderSparkMaxImpl::Set(speed);
}



/**************************************************************/
// ThunderSMTalonImpl

ThunderSMTalonImpl::ThunderSMTalonImpl(int canid) : ThunderSparkMaxImpl("CAN Talon", canid), talon(canid)
{
    talon.ConfigSelectedFeedbackSensor(FeedbackDevice::CTRE_MagEncoder_Relative, 0, 0);
}

void ThunderSMTalonImpl::Set(double speed)
{
    talon.Set(ControlMode::PercentOutput, speed);
    ThunderSparkMaxImpl::Set(speed);
}

void ThunderSMTalonImpl::SetOpenLoopRampRate(double rate)
{
    talon.ConfigOpenloopRamp(rate);
}

double ThunderSMTalonImpl::GetEncoder()
{
    int n = talon.GetSelectedSensorPosition(0);
    double revs = ((double)n) / 4096;
    return revs;
}

void ThunderSMTalonImpl::SetEncoder(double rotations)
{
    int n = (int)(rotations * 4096);
    talon.SetSelectedSensorPosition(n, 0, 0);
}

void ThunderSMTalonImpl::SetIdleMode(IdleMode idleMode)
{
    talon.SetNeutralMode(idleMode == IdleMode::BRAKE ? NeutralMode::Brake : NeutralMode::Coast);
}

void ThunderSMTalonImpl::SetInverted(bool invert)
{
    talon.SetInverted(invert ? ctre::phoenix::motorcontrol::InvertType::InvertMotorOutput : ctre::phoenix::motorcontrol::InvertType::None);
}






/**************************************************************/
// ThunderSMTalonImpl

ThunderSMCANImpl::ThunderSMCANImpl(int canid) : ThunderSparkMaxImpl("CAN SparkMax", canid),
        spark(canid, rev::CANSparkMaxLowLevel::MotorType::kBrushless),
        enc(spark.GetEncoder()),
        pid(spark.GetPIDController())
{
}

void ThunderSMCANImpl::Set(double speed)
{
    spark.Set(speed);
    ThunderSparkMaxImpl::Set(speed);
}

void ThunderSMCANImpl::SetOpenLoopRampRate(double rate)
{
    spark.SetOpenLoopRampRate(rate);
}

double ThunderSMCANImpl::GetEncoder()
{
    return enc.GetPosition();
}

double ThunderSMCANImpl::Get() {
    return spark.GetAppliedOutput();
}

void ThunderSMCANImpl::SetEncoder(double rotations)
{
    enc.SetPosition(rotations);
}

void ThunderSMCANImpl::SetIdleMode(IdleMode idleMode)
{
    spark.SetIdleMode(idleMode == IdleMode::BRAKE ? rev::CANSparkMax::IdleMode::kBrake : rev::CANSparkMax::IdleMode::kCoast);
}

void ThunderSMCANImpl::SetInverted(bool invert)
{
    spark.SetInverted(invert);
}

void ThunderSMCANImpl::Follow(ThunderSparkMax *leader, bool invertOutput)
{
    ThunderSparkMaxImpl *impl = static_cast<ThunderSparkMaxImpl *>(leader);
    rev::CANSparkMax *lsm = impl->getSparkMax();
    if (lsm) {
        spark.Follow(*lsm, invertOutput);
    } else {
        printf("ERROR: CAN id %d cannot follow a fake spark max!\n", spark.GetDeviceId());
    }
}


/**************************************************************/
// ThunderSMJagImpl

ThunderSMPWMImpl::ThunderSMPWMImpl(int pwmChannel) : ThunderSparkMaxImpl("PWM SparkMax", pwmChannel), spark(pwmChannel)
{
}

void ThunderSMPWMImpl::Set(double speed)
{
    spark.Set(speed);
    ThunderSparkMaxImpl::Set(speed);
}

